(define pause
  (lambda ()
    (display "Press any Enter...")
    (read-char)
    (display "")))

; purely functional approach
(define same-fringe? 
  (lambda (tree1 tree2) 
    (let loop ((ftree1 (flatten tree1)) 
               (ftree2 (flatten tree2))) 
      (cond ((and (null? ftree1) (null? ftree2)) #t) 
            ((or (null? ftree1) (null? ftree2)) #f) 
            ((eqv? (car ftree1) (car ftree2)) 
             (loop (cdr ftree1) (cdr ftree2))) 
            (else #f))))) 
 
(define flatten 
  (lambda (tree) 
    (cond ((null? tree) '()) 
          ((pair? (car tree)) 
           (append (flatten (car tree)) 
                   (flatten (cdr tree)))) 
          (else 
           (cons (car tree) 
                 (flatten (cdr tree)))))))

; with call/cc
(define tree->generator 
  (lambda (tree) 
    (let ((caller '*)) 
      (letrec 
          ((generate-leaves 
            (lambda () 
              (let loop ((tree tree)) 
                (cond ((null? tree) 'skip) 
                      ((pair? tree) 
                       (loop (car tree)) 
                       (loop (cdr tree))) 
                      (else 
                       (call/cc 
                        (lambda (rest-of-tree) 
                          (set! generate-leaves 
                            (lambda () 
                              (rest-of-tree 'resume))) 
                          (caller tree)))))) 
              (caller '())))) 
        (lambda () 
          (call/cc 
           (lambda (k) 
             (set! caller k) 
             (generate-leaves))))))))

(define same-fringe/cc? 
  (lambda (tree1 tree2) 
    (let ((gen1 (tree->generator tree1)) 
          (gen2 (tree->generator tree2))) 
      (let loop () 
        (let ((leaf1 (gen1)) 
              (leaf2 (gen2))) 
          (if (eqv? leaf1 leaf2) 
              (if (null? leaf1) #t (loop)) 
              #f))))))

; with coroutines
(define-macro coroutine 
  (lambda (x . body) 
    `(letrec ((local-control-state 
               (lambda (,x) ,@body)) 
              (resume 
               (lambda (c v) 
                 (call/cc 
                  (lambda (k) 
                    (set! local-control-state k) 
                    (c v)))))) 
       (lambda (v) 
         (local-control-state v))))) 

(define make-matcher-coroutine 
  (lambda (tree-cor-1 tree-cor-2) 
    (coroutine dont-need-an-init-arg 
      (let loop () 
        (let ((leaf1 (resume tree-cor-1 'get-a-leaf)) 
              (leaf2 (resume tree-cor-2 'get-a-leaf))) 
          (if (eqv? leaf1 leaf2) 
              (if (null? leaf1) #t (loop)) 
              #f)))))) 

(define make-leaf-gen-coroutine 
  (lambda (tree matcher-cor) 
    (coroutine dont-need-an-init-arg 
      (let loop ((tree tree)) 
        (cond ((null? tree) 'skip) 
              ((pair? tree) 
               (loop (car tree)) 
               (loop (cdr tree))) 
              (else 
               (resume matcher-cor tree)))) 
      (resume matcher-cor '())))) 

(define same-fringe-cr? 
  (lambda (tree1 tree2) 
    (letrec ((tree-cor-1 
              (make-leaf-gen-coroutine 
               tree1 
               (lambda (v) (matcher-cor v)))) 
             (tree-cor-2 
              (make-leaf-gen-coroutine 
               tree2 
               (lambda (v) (matcher-cor v)))) 
             (matcher-cor 
              (make-matcher-coroutine 
               (lambda (v) (tree-cor-1 v)) 
               (lambda (v) (tree-cor-2 v))))) 
      (matcher-cor 'start-ball-rolling)))) 

(same-fringe? '(1 (2 3)) '((1 2) 3)) 
(same-fringe? '(1 2 3) '(1 (3 2))) 

(pause)

(same-fringe/cc? '(1 (2 3)) '((1 2) 3)) 
(same-fringe/cc? '(1 2 3) '(1 (3 2))) 

(pause)

(same-fringe-cr? '(1 (2 3)) '((1 2) 3)) 
(same-fringe-cr? '(1 2 3) '(1 (3 2))) 
