(load-relative "../2. alkalom/defstruct.scm")

(defstruct standard-class 
  slots superclass method-names method-vector)

(define-macro create-class 
  (lambda (superclass slots . methods) 
    `(create-class-proc 
      ,superclass 
      (list ,@(map (lambda (slot) `',slot) slots)) 
      (list ,@(map (lambda (method) `',(car method)) methods)) 
      (vector ,@(map (lambda (method) `,(cadr method)) methods)))))

(define make-instance 
  (lambda (class . slot-value-twosomes) 
 
    ;Find `n', the number of slots in `class'. 
    ;Create an instance vector of length `n + 1', 
    ;because we need one extra element in the instance 
    ;to contain the class. 
 
    (let* ((slotlist (standard-class.slots class)) 
           (n (length slotlist)) 
           (instance (make-vector (+ n 1)))) 
      (vector-set! instance 0 class) 
 
      ;Fill each of the slots in the instance 
      ;with the value as specified in the call to 
      ;`make-instance'. 
 
      (let loop ((slot-value-twosomes slot-value-twosomes)) 
        (if (null? slot-value-twosomes) instance 
            (let ((k (list-position (car slot-value-twosomes)  
                                    slotlist))) 
              (vector-set! instance (+ k 1)  
                (cadr slot-value-twosomes)) 
              (loop (cddr slot-value-twosomes))))))))

(define class-of 
  (lambda (x) 
    (if (vector? x) 
        (let ((n (vector-length x))) 
          (if (>= n 1) 
              (let ((c (vector-ref x 0))) 
                (if (standard-class? c) c #t)) 
              #t)) 
        #t)))

(define slot-value 
  (lambda (instance slot) 
    (let* ((class (class-of instance)) 
           (slot-index 
            (list-position slot (standard-class.slots class)))) 
      (vector-ref instance (+ slot-index 1))))) 
 
(define set!slot-value 
  (lambda (instance slot new-val) 
    (let* ((class (class-of instance)) 
           (slot-index 
            (list-position slot (standard-class.slots class)))) 
      (vector-set! instance (+ slot-index 1) new-val))))

(define create-class-proc 
  (lambda (superclass slots method-names method-vector) 
    (make-standard-class 
     'superclass superclass 
     'slots 
     (let ((superclass-slots  
            (if (not (eqv? superclass #t)) 
                (standard-class.slots superclass) 
                '()))) 
       (if (null? superclass-slots) slots 
           (delete-duplicates 
            (append slots superclass-slots)))) 
     'method-names method-names 
     'method-vector method-vector))) 

(define delete-duplicates 
  (lambda (s) 
    (if (null? s) s 
        (let ((a (car s)) (d (cdr s))) 
          (if (memv a d) (delete-duplicates d) 
              (cons a (delete-duplicates d)))))))

(define send 
  (lambda (method instance . args) 
    (let ((proc 
           (let loop ((class (class-of instance))) 
             (if (eqv? class #t) (error 'send) 
                 (let ((k (list-position  
                           method 
                           (standard-class.method-names class)))) 
                   (if k 
                       (vector-ref (standard-class.method-vector class) k) 
                       (loop (standard-class.superclass class)))))))) 
      (apply proc instance args))))

