;; Programmer: John Stone, Grinnell College.
;; June 9, 1996.

(define GPS
  (lambda (initial-state goals operations)
    (let ((result (achieve-all goals initial-state operations)))
      (if result
          (display-steps (cdr result))
          (writeln "GPS was unable to find a solution.")))))

(define achieve-all
  (lambda (goals initial-state operations)
    (if (null? goals)
        (cons initial-state '())
        (let ((first-part (achieve (car goals) initial-state operations)))
          (and first-part
               (let ((rest-part
                      (achieve-all (cdr goals) (car first-part) operations)))
                 (and rest-part
                      (cons (car rest-part)
                            (append (cdr first-part) (cdr rest-part))))))))))

(define achieve
  (lambda (goal initial-state operations)
    (if (member? goal initial-state)
        (cons initial-state '())
        (try (lambda (possible)
               (let ((result (achieve-all (preconditions possible)
                                          initial-state
                                          operations)))
                 (and result
                      (cons (apply-operation (car result) possible)
                            (attach-at-end possible (cdr result))))))
             (filter (lambda (op)
                       (member? goal (products op)))
                     operations)))))

(define member?
  (lambda (val li)
    (cond ((null? li) #f)
          ((equal? val (car li)) #t)
          (else (member? val (cdr li))))))

(define try
  (lambda (proc li)
    (if (null? li)
        #f
        (or (proc (car li))
            (try proc (cdr li))))))

(define filter
  (lambda (pred li)
    (letrec ((helper
              (lambda (rest so-far)
                (if (null? rest)
                    (reverse so-far)
                    (helper (cdr rest)
                            (let ((first (car rest)))
                              (if (pred first)
                                  (cons first so-far)
                                  so-far)))))))
      (helper li '()))))

(define attach-at-end
  (lambda (val li)
    (if (null? li)
        (cons val '())
        (cons (car li) (attach-at-end val (cdr li))))))

(define apply-operation
  (lambda (state operation)
    (union (products operation)
           (set-difference state (sumpta operation)))))

(define union
  (lambda (set-1 set-2)
    (letrec ((helper
              (lambda (set so-far)
                (if (null? set)
                    so-far
                    (helper (cdr set)
                            (let ((first (car set)))
                              (if (member first set-2)
                                  so-far
                                  (cons first so-far))))))))
      (helper set-1 set-2))))

(define set-difference
  (lambda (set-1 set-2)
    (letrec ((helper
              (lambda (set so-far)
                (if (null? set)
                    so-far
                    (helper (cdr set)
                            (let ((first (car set)))
                              (if (member first set-2)
                                  so-far
                                  (cons first so-far))))))))
      (helper set-1 '()))))

(define make-op 
  (lambda (action preconditions products sumpta)
    (list action preconditions products sumpta)))

(define action car)
(define preconditions cadr)
(define products caddr)
(define sumpta cadddr)

(define writeln
  (lambda args
    (for-each display args)
    (newline)))

(define display-steps
  (lambda (operation-sequence)
    (for-each (lambda (operation)
                (writeln (action operation)))
              operation-sequence)))

(define *school-ops*
  (list

   ;; If your son is at home and your car works, it is possible to drive
   ;; him to school.  (Then he'll be at school and will no longer be at
   ;; home.) 

   (make-op "drive son to school"
            '(son-at-home car-works)
            '(son-at-school)
            '(son-at-home))

   ;; If your car needs a new battery, and the mechanic knows the problem
   ;; and has been paid, it is possible him to install the new battery.
   ;; Then the car will work.

   (make-op "have the mechanic install a new battery"
            '(car-needs-battery mechanic-knows-problem
                                mechanic-has-money) 
            '(car-works)
            '(car-needs-battery))

   ;; If you can communicate with the mechanic, you can tell him about the
   ;; problem with your car, and then he'll know what it is.

   (make-op "tell the mechanic what the problem is"
            '(in-communication-with-mechanic)
            '(mechanic-knows-problem)
            '())

   ;; If you know the mechanic's telephone number, you can call him, and
   ;; then you'll be able to communicate with him.

   (make-op "telephone the mechanic"
            '(know-phone-number)
            '(in-communication-with-mechanic)
            '())

   ;; If you have a telephone book, you can look up the mechanic's number,
   ;; and then you'll know what it is. 

   (make-op "look up the telephone number"
            '(have-phone-book)
            '(know-phone-number)
            '())

   ;; If you have money, then you can pay the mechanic.  Then he'll have
   ;; the money and you won't. 

   (make-op "pay the mechanic"
            '(have-money)
            '(mechanic-has-money)
            '(have-money))))
