;; Programmer: John Stone, Grinnell College.
;; Original version: September 20 -- October 8, 1991.
;; Ported to Scheme: June 14 -- 21, 1996.

;(require 'random)

(define penny-matcher
  (lambda ()
    (let loop ((round 1)
               (my-wins 0)
               (your-wins 0)
               (transcript (initialize-table))
               (last-move '())
               (last-move-but-one '()))

      (if (< 100 round)

          ; Game's over; announce the outcome.
          (report-results my-wins your-wins)

          ; Choose a move and get the player's move simultaneously.
          (let ((my-move (select-my-move round
                                         last-move
                                         last-move-but-one
                                         transcript))
                (your-move (select-your-move round)))

            ; Announce the outcome of the current round.
            (report-round my-move your-move my-wins
                          your-wins)

            ; On to the next round, then.
            (loop (+ round 1)
                  (if (eq? my-move your-move)
                      (+ my-wins 1)
                      my-wins)
                  (if (eq? my-move your-move)
                      your-wins
                      (+ 1 your-wins))
                  (if (<= round 2)   ; Start recording behavior
                                     ;   only after round 2.
                      transcript
                      (update-table transcript
                                    (list last-move
                                          last-move-but-one)
                                    your-move))
                  (cons my-move your-move)
                  last-move))))))

(define initialize-table
  (lambda ()
    (let* ((faces '(heads tails))
           (moves (apply append
                         (map (lambda (face-1)
                                (map (lambda (face-2)
                                       (cons face-1 face-2))
                                     faces))
                              faces)))
           (situations (apply append
                             (map (lambda (move-1)
                                    (map (lambda (move-2)
                                           (list move-1 move-2))
                                         moves))
                                  moves))))
    (map (lambda (situation)
           (cons situation (cons 0 0)))
         situations))))

(define report-results
  (lambda (my-wins your-wins)
    (cond ((< my-wins your-wins)
           (display "Congratulations -- you outwitted me.")
           (newline)
           (display "I owe you ")
           (display (- your-wins my-wins))
           (display " cents.")
           (newline))
          ((< your-wins my-wins)
           (display "Hmm.  Looks like you owe me ")
           (display (- my-wins your-wins))
           (display " cents.")
           (newline)
           (display "Keep it -- there's no way for me to spend it anyway.")
           (newline))
          (else
           (display "Hmm.  Well, that was pointless.")
           (newline)
           (display "I guess I need a better cognitive model!")
           (newline)))
    (newline)))

(define select-my-move
  (lambda (round last-move last-move-but-one transcript)
    (if (<= round 2)
        (toss-a-coin)
        (let ((past-record (lookup (list last-move
                                         last-move-but-one)
                                   transcript)))
          (let ((head-plays (car past-record))
                (tail-plays (cdr past-record)))
            (cond ((< head-plays tail-plays) 'tails)
                  ((< tail-plays head-plays) 'heads)
                  (else (toss-a-coin))))))))

(define toss-a-coin
  (lambda ()
    (if (zero? (random 2))
        'heads
        'tails)))

(define lookup
  (lambda (situation transcript)
    (let lookup-it ((rest transcript))
      (if (equal? situation (caar rest))
          (cdar rest)
          (lookup-it (cdr rest))))))

(define select-your-move
  (lambda (round-number)
    (display "Round #")
    (display round-number)
    (newline)
    (display "Your move (heads or tails): ")
    (let loop ((input (read)))
      (consume-line)
      (if (or (eq? input 'heads)
              (eq? input 'tails))
          input
          (begin
            (display "Type `heads' or `tails', please.")
            (newline)
            (display "Your move: ")
            (loop (read)))))))

(define consume-line
  (lambda ()
    (let ((next-char (peek-char)))
      (or (eof-object? next-char)
          (begin
            (read-char)
            (or (eq? next-char #\newline)
                (consume-line)))))))

(define report-round
  (lambda (my-move your-move my-wins your-wins)
    (display "You showed ")
    (display your-move)
    (display ", I showed ")
    (display my-move)
    (display ".")
    (newline)
    (if (eq? my-move your-move)
        (begin
          (display "They match, so I win this round.")
          (newline)
          (display "That's ")
          (display (+ my-wins 1))
          (display " for me and ")
          (display your-wins)
          (display " for you.")
          (newline))
        (begin
          (display "They don't match, so you win this round.")
          (newline)
          (display "That's ")
          (display my-wins)
          (display " for me and ")
          (display (+ your-wins 1))
          (display " for you.")
          (newline)))
    (newline)))

(define update-table
  (lambda (transcript situation your-move)
    (let loop ((rest transcript)
               (passed '()))
      (if (equal? situation (caar rest))
          (cons (cons situation
                      (if (eq? your-move 'heads)
                          (cons (+ (cadar rest) 1)
                                (cddar rest))
                          (cons (cadar rest)
                                (+ (cddar rest) 1))))
                (append (reverse passed)
                        (cdr rest)))
          (loop (cdr rest) (cons (car rest) passed))))))
